#include <Bounce.h>
#include <ResponsiveAnalogRead.h>
#include <Adafruit_NeoPixel.h>
#include <MIDI.h>


MIDI_CREATE_INSTANCE(HardwareSerial, Serial1, MIDI1);
MIDI_CREATE_INSTANCE(HardwareSerial, Serial2, MIDI2);
MIDI_CREATE_INSTANCE(HardwareSerial, Serial3, MIDI3);
MIDI_CREATE_INSTANCE(HardwareSerial, Serial4, MIDI4);


int buttonA1Pin = 24;
int buttonA2Pin = 25;
int buttonB1Pin = 26;
int buttonB2Pin = 27;
int buttonC1Pin = 28;
int buttonC2Pin = 29;
int buttonD1Pin = 30;
int buttonD2Pin = 31;

ResponsiveAnalogRead knobA1(14, true, 1);
ResponsiveAnalogRead knobA2(15, true, 1);
ResponsiveAnalogRead knobB1(16, true, 1);
ResponsiveAnalogRead knobB2(17, true, 1);
ResponsiveAnalogRead knobC1(18, true, 1);
ResponsiveAnalogRead knobC2(19, true, 1);
ResponsiveAnalogRead knobD1(20, true, 1);
ResponsiveAnalogRead knobD2(21, true, 1);

Adafruit_NeoPixel pixelA = Adafruit_NeoPixel(8, 33, NEO_RGB + NEO_KHZ800);
Adafruit_NeoPixel pixelB = Adafruit_NeoPixel(8, 34, NEO_RGB + NEO_KHZ800);
Adafruit_NeoPixel pixelC = Adafruit_NeoPixel(8, 35, NEO_RGB + NEO_KHZ800);
Adafruit_NeoPixel pixelD = Adafruit_NeoPixel(8, 36, NEO_RGB + NEO_KHZ800);

Bounce buttonA1 = Bounce(buttonA1Pin, 5);
Bounce buttonA2 = Bounce(buttonA2Pin, 5);
Bounce buttonB1 = Bounce(buttonB1Pin, 5);
Bounce buttonB2 = Bounce(buttonB2Pin, 5);
Bounce buttonC1 = Bounce(buttonC1Pin, 5);
Bounce buttonC2 = Bounce(buttonC2Pin, 5);
Bounce buttonD1 = Bounce(buttonD1Pin, 5);
Bounce buttonD2 = Bounce(buttonD2Pin, 5);



// ------------------------------------------------------------

byte singleLed[] = {B00000001, B00000010, B00000100, B00001000, B00010000, B00100000, B01000000, B10000000};
byte filledLed[] = {B00000001, B00000011, B00000111, B00001111, B00011111, B00111111, B01111111, B11111111};
byte syncLed[] = {B00011000, B00100100, B01000010, B10000001};
byte knightRider[] = {B00000001, B00000010, B00000100, B00001000, B00010000, B00100000, B01000000, B10000000,
                      B10000000, B01000000, B00100000, B00010000, B00001000, B00000100, B00000010, B00000001
                     };




// ------------------------------------------------------------

int unit_ID = 0;
int mode = -1;
int pos = 0;

const int posMult = 4;    //the position value is multiplied in Max for blinking effects
int Brightness = 127;
int knightPos = 0;
unsigned long knightMillis = 0;
unsigned long sendlimitMillis = 0;

int t = 0;
int cR = 0;
int cG = 0;
int cB = 0;

bool sendValues = false;





// ------------------------------------------------------------

void setup()   {
  Serial.begin(9600);

  MIDI1.begin(MIDI_CHANNEL_OMNI);
  MIDI2.begin(MIDI_CHANNEL_OMNI);
  MIDI3.begin(MIDI_CHANNEL_OMNI);
  MIDI4.begin(MIDI_CHANNEL_OMNI);

  pinMode(buttonA1Pin, INPUT_PULLUP);
  pinMode(buttonA2Pin, INPUT_PULLUP);
  pinMode(buttonB1Pin, INPUT_PULLUP);
  pinMode(buttonB2Pin, INPUT_PULLUP);
  pinMode(buttonC1Pin, INPUT_PULLUP);
  pinMode(buttonC2Pin, INPUT_PULLUP);
  pinMode(buttonD1Pin, INPUT_PULLUP);
  pinMode(buttonD2Pin, INPUT_PULLUP);

  pixelA.begin();
  pixelB.begin();
  pixelC.begin();
  pixelD.begin();

}



// ---- LOOP START ----------------------------------------------


void loop() {

  handleMidi();

  if (Serial.available() > 2 ) {

    unit_ID = (char)Serial.read();   //  From Max, neoPixel led ID
    mode = (char)Serial.read();      //  From Max, mode
    pos = (char)Serial.read();       //  From Max, position

    switch (unit_ID) {
      case 10:
        sendValues = false;
        break;
      case 11:
        sendValues = true;
        break;
      case 20:
        Brightness = mode;
        break;
      case 99:
        _reboot_Teensyduino_();
        break;
    }

  }

  handleMidi();

  knobA1.update();
  knobA2.update();
  buttonA1.update();
  buttonA2.update();

  handleMidi();

  knobB1.update();
  knobB2.update();
  buttonB1.update();
  buttonB2.update();

  handleMidi();

  knobC1.update();
  knobC2.update();
  buttonC1.update();
  buttonC2.update();

  handleMidi();

  knobD1.update();
  knobD2.update();
  buttonD1.update();
  buttonD2.update();

  handleMidi();

  // ------------------------------------------------------------

  if (millis() - sendlimitMillis > 5) {  //  TEMP time limit test on knob value
    if (knobA1.hasChanged()) {
      writeValue("knobA1", knobA1.getValue());
      handleMidi();
    }
    if (knobA2.hasChanged()) {
      writeValue("knobA2", knobA2.getValue());
      handleMidi();
    }
    if (knobB1.hasChanged()) {
      writeValue("knobB1", knobB1.getValue());
      handleMidi();
    }
    if (knobB2.hasChanged()) {
      writeValue("knobB2", knobB2.getValue());
      handleMidi();
    }
    if (knobC1.hasChanged()) {
      writeValue("knobC1", knobC1.getValue());
      handleMidi();
    }
    if (knobC2.hasChanged()) {
      writeValue("knobC2", knobC2.getValue());
      handleMidi();
    }
    if (knobD1.hasChanged()) {
      writeValue("knobD1", knobD1.getValue());
      handleMidi();
    }
    if (knobD2.hasChanged()) {
      writeValue("knobD2", knobD2.getValue());
      handleMidi();
    }

    sendlimitMillis = millis();

  }




  // ------------------------------------------------------------

  if (buttonA1.risingEdge()) {
    writeValue("buttonA1", 0);
    handleMidi();
  }
  if (buttonA1.fallingEdge()) {
    writeValue("buttonA1", 1);
    handleMidi();
  }
  if (buttonA2.risingEdge()) {
    writeValue("buttonA2", 0);
    handleMidi();
  }
  if (buttonA2.fallingEdge()) {
    writeValue("buttonA2", 1);
    handleMidi();
  }
  if (buttonB1.risingEdge()) {
    writeValue("buttonB1", 0);
    handleMidi();
  }
  if (buttonB1.fallingEdge()) {
    writeValue("buttonB1", 1);
    handleMidi();
  }
  if (buttonB2.risingEdge()) {
    writeValue("buttonB2", 0);
    handleMidi();
  }
  if (buttonB2.fallingEdge()) {
    writeValue("buttonB2", 1);
    handleMidi();
  }
  if (buttonC1.risingEdge()) {
    writeValue("buttonC1", 0);
    handleMidi();
  }
  if (buttonC1.fallingEdge()) {
    writeValue("buttonC1", 1);
    handleMidi();
  }
  if (buttonC2.risingEdge()) {
    writeValue("buttonC2", 0);
    handleMidi();
  }
  if (buttonC2.fallingEdge()) {
    writeValue("buttonC2", 1);
    handleMidi();
  }
  if (buttonD1.risingEdge()) {
    writeValue("buttonD1", 0);
    handleMidi();
  }
  if (buttonD1.fallingEdge()) {
    writeValue("buttonD1", 1);
    handleMidi();
  }
  if (buttonD2.risingEdge()) {
    writeValue("buttonD2", 0);
    handleMidi();
  }
  if (buttonD2.fallingEdge()) {
    writeValue("buttonD2", 1);
    handleMidi();
  }




  // ------------------------------------------------------------

  switch (mode) {
    case 1:                                  //OFF

        updateNeoPixel8(filledLed[pos / posMult], 0, 0, 0); //rgb off
        handleMidi();


      break;

    case 2:                                  //PLAY

      updateNeoPixel8(filledLed[pos / posMult], 0, Brightness, 0);  // RGB green
      handleMidi();


      break;

    case 3:                                  //REC
      if (pos % posMult <= 1) {

        updateNeoPixel8(filledLed[pos / posMult], Brightness, 0, 0);  // RGB red
        handleMidi();
      }

      break;

    case 4:                                   //RESYNC

      updateNeoPixel8(filledLed[7], (pos % 8 * Brightness / 8), (pos % 8 * Brightness / 8) / 2, 0); // RGB soft flash yellow
      handleMidi();

      break;

    case 5:                                    //CUE

      if (pos % 2 == 1) {

        updateNeoPixel8(singleLed[pos / posMult], Brightness, 0, 0);  // RGB red
        handleMidi();
      } else {

        updateNeoPixel8(filledLed[pos / posMult], 0, 0, 0); //rgb off
        handleMidi();
      }

      break;

    case 6:                                     //FLASH  (Play but yellow)

      updateNeoPixel8(filledLed[pos / posMult], Brightness, Brightness, 0);  // RGB yellow
      handleMidi();

      break;

    case 7:                                  //NIKLAS RAINBOW MODE

      t++;
      cR = 128 + sin(t / 40) * 128;
      cG = 128 + sin(t / 67) * 128;
      cB = 128 + sin(t / 101) * 128;

      updateNeoPixel8(singleLed[pos / posMult], cR, cG, cB);  // RGB green
      handleMidi();

      break;

    default:
      //      updateLeds(knightRider[knightPos]);     //knight rider mode on startup
      updateNeoPixel8(knightRider[knightPos], Brightness, 0, 0);
      if (millis() - knightMillis > 50) {
        knightPos = (knightPos + 1) % 16;
        knightMillis = millis();
      }
      handleMidi();
      break;
  }

}


// ---- LOOP END ------------------------------------------------



void updateNeoPixel8(byte value, int r, int g, int b) {
  if ((unit_ID == 1) || (mode == -1)) {
    for (int i = 0; i <= 7; i++) {
      pixelA.setPixelColor(i, pixelA.Color((bitRead(value, i)) * r, (bitRead(value, i)) * g, (bitRead(value, i)) * b));
      pixelA.show();        // This sends the updated pixel color to the hardware.
    }
  }
  if ((unit_ID == 2) || (mode == -1)) {
    for (int i = 0; i <= 7; i++) {
      pixelB.setPixelColor(i, pixelB.Color((bitRead(value, i)) * r, (bitRead(value, i)) * g, (bitRead(value, i)) * b));
      pixelB.show();        // This sends the updated pixel color to the hardware.
    }
  }
  if ((unit_ID == 3) || (mode == -1)) {
    for (int i = 0; i <= 7; i++) {
      pixelC.setPixelColor(i, pixelC.Color((bitRead(value, i)) * r, (bitRead(value, i)) * g, (bitRead(value, i)) * b));
      pixelC.show();        // This sends the updated pixel color to the hardware.
    }
  }
  if ((unit_ID == 4) || (mode == -1)) {
    for (int i = 0; i <= 7; i++) {
      pixelD.setPixelColor(i, pixelD.Color((bitRead(value, i)) * r, (bitRead(value, i)) * g, (bitRead(value, i)) * b));
      pixelD.show();        // This sends the updated pixel color to the hardware.
    }
  }
}

void writeValue(String nm, int value) {
  if (sendValues == true) {
    Serial.print(nm);
    Serial.print(" ");
    Serial.println(value);
  }

}

void handleMidi() {
  if (usbMIDI.read()) {
    // get the USB MIDI message, defined by these 5 numbers (except SysEX)
    byte type = usbMIDI.getType();
    byte channel = usbMIDI.getChannel();
    byte data1 = usbMIDI.getData1();
    byte data2 = usbMIDI.getData2();
    byte cable = usbMIDI.getCable();

    // forward this message to 1 of the 3 Serial MIDI OUT ports
    if (type != usbMIDI.SystemExclusive) {
      // Normal messages, first we must convert usbMIDI's type (an ordinary
      // byte) to the MIDI library's special MidiType.
      midi::MidiType mtype = (midi::MidiType)type;

      // Then simply give the data to the MIDI library send()
      switch (cable) {
        case 0:
          MIDI1.send(mtype, data1, data2, channel);
          break;
        case 1:
          MIDI2.send(mtype, data1, data2, channel);
          break;
        case 2:
          MIDI3.send(mtype, data1, data2, channel);
          break;
        case 3:
          MIDI4.send(mtype, data1, data2, channel);
          break;
      }

    } else {
      // SysEx messages are special.  The message length is given in data1 & data2
      unsigned int SysExLength = data1 + data2 * 256;
      switch (cable) {
        case 0:
          MIDI1.sendSysEx(SysExLength, usbMIDI.getSysExArray(), true);
          break;
        case 1:
          MIDI2.sendSysEx(SysExLength, usbMIDI.getSysExArray(), true);
          break;
        case 2:
          MIDI3.sendSysEx(SysExLength, usbMIDI.getSysExArray(), true);
          break;
        case 3:
          MIDI4.sendSysEx(SysExLength, usbMIDI.getSysExArray(), true);
          break;
      }
    }
  }
}
